unit AdvancedIcon;

interface

uses
  Windows, SysUtils, Classes, Graphics, IconTypes, IconImage;

type
  TAdvancedIcon = class;
  TIconImageList = class;

  TAdvancedIcon = class
  private
    fImages : TIconImageList;
  protected
    function LoadHeader(Stream: TStream; var IconCount: word): boolean;
  public
    function LoadFromFile(Filename: string): boolean;
    function LoadFromResource(Instance: THandle; ResourceName: pChar): boolean;
    function LoadFromResourceID(Instance: THandle; ResourceID: integer): boolean;
    function LoadFromResourceName(Instance: THandle; ResourceName: string): boolean;

    procedure SaveToStream(Stream: TStream);
    procedure SaveToFile(Filename: string);
    procedure Assign(Source: TAdvancedIcon); virtual;

    constructor Create; virtual;
    destructor Destroy; override;

    property Images: TIconImageList read fImages;
  end;

  TIconImageList = class
  private
    fList: PPointerList;
    fCount: word;
    fCapacity: word;
    function GetIconImage(Index: word): TIconImage;
    procedure SetCapacity(Value: word);
  protected
    procedure DeleteImage(Index: word; FreeImage: boolean);
  public
    procedure Add(IconImage: TIconImage);
    procedure Delete(Index: word);
    procedure AddFromIconImageList(IconSource: TAdvancedIcon; Index: word);
    procedure Clear;
    procedure SortByResolution;
    procedure SortByColor;
    constructor Create;
    destructor Destroy; override;
    property Count: word read fCount;
    property Image[Index: word]: TIconImage read GetIconImage;
    property Capacity: word read fCapacity write SetCapacity;
  end;


implementation

function IconImageColorSort(Item1, Item2: TIconImage): integer;
Var
  A,B: integer;
begin
  A:=Item1.Info.BitCount;
  B:=Item2.Info.BitCount;
  if A = B then
    begin
      A:=Item1.Info.Width;
      B:=Item2.Info.Width;
      if A = B then
        begin
         Result:=0
        end
      else
        Result:=(2*integer(A>B))-1
    end
  else
    Result:=(2*integer(A>B))-1
end;

function IconImageResolutionSort(Item1, Item2: TIconImage): integer;
var
  A,B: integer;
begin
  A:=Item1.Info.Width;
  B:=Item2.Info.Width;
  if A = B then
    begin
      A:=Item1.Info.BitCount;
      B:=Item2.Info.BitCount;
      if A = B then
        begin
          Result:=0
        end
      else
        Result:=(2*integer(A>B))-1
    end
  else
    Result:=(2*integer(A>B))-1
end;

procedure QuickSort(SortList: PPointerList; L, R: integer; SCompare: TListSortCompare);
var
  I,J: integer;
  P,T: Pointer;
begin
  repeat
    I:=L; J:=R;
    P:=SortList^[(L + R) shr 1];
    repeat
      while SCompare(SortList^[I], P) < 0 do inc(I);
      while SCompare(SortList^[J], P) > 0 do dec(J);
      if I <= J then
        begin
          T:=SortList^[I];
          SortList^[I]:=SortList^[J];
          SortList^[J]:=T;
          inc(I); dec(J)
        end
    until I > J;
    if L < J then QuickSort(SortList, L, J, SCompare);
    L:=I
  until I >= R
end;

procedure TIconImageList.SortByResolution;
begin
  QuickSort(fList,0,Count-1,TListSortCompare(@IconImageResolutionSort))
end;

procedure TIconImageList.SortByColor;
begin
  QuickSort(fList,0,Count-1,TListSortCompare(@IconImageColorSort))
end;

constructor TAdvancedIcon.Create;
begin
  fImages:=TIconImageList.Create
end;

destructor TAdvancedIcon.Destroy;
begin
  fImages.Free
end;

constructor TIconImageList.Create;
begin
  fList:=nil;
  fCount:=0; fCapacity:=0
end;

procedure TIconImageList.SetCapacity(Value: word);
begin
  if Value < fCount then Value:=fCount;
  ReallocMem(FList,Value * SizeOf(Pointer));
  fCapacity:=Value
end;

procedure TIconImageList.Add(IconImage: TIconImage);
begin
  inc(fCount);
  if fCount>fCapacity then SetCapacity(fCount);
  fList^[fCount-1]:=IconImage
end;

procedure TIconImageList.Delete(Index: word);
begin
  DeleteImage(Index,True)
end;

procedure TIconImagelist.DeleteImage(Index: word; FreeImage: boolean);
begin
  if Index >= fCount then exit;
  if FreeImage then TIconImage(fList^[Index]).Free;
  dec(fCount);  if Index < fCount then System.Move(fList^[Index + 1], fList^[Index],(fCount - Index) * SizeOf(Pointer))
end;

procedure TIconImageList.AddFromIconImageList(IconSource: TAdvancedIcon; Index: word);
var
  Image: TIconImage;
begin
  if not Assigned(IconSource) then exit;
  Image:=IconSource.Images.Image[Index];
  if not Assigned(Image) then exit;
  IconSource.Images.DeleteImage(Index,False);
  Add(Image)
end;

function TIconImageList.GetIconImage(Index: word): TIconImage;
begin
  Result:=nil;
  if Index > fCapacity then exit;
  Result:=fList^[Index]
end;

procedure TIconImageList.Clear;
var
  I: integer;
begin
  if Assigned(fList) then
    begin
      for I:=1 to fCount do TIconImage(fList^[I-1]).free;
      Freemem(fList);
      fList:=nil
    end;
  fCapacity:=0; fCount:=0
end;

destructor TIconImageList.Destroy;
begin
  Clear;
  inherited
end;

procedure TAdvancedIcon.Assign(Source: TAdvancedIcon);
var
  I: Integer;
  SourceImage,DestImage: TIconImage;
begin
  Images.Clear;
  Images.Capacity:=Source.Images.Count;
  for I:=1 to Source.Images.Count do
    begin
      SourceImage:=Source.Images.Image[I-1];
      if Assigned(SourceImage) then
        begin
          DestImage:=TIconImage.CreateCopy(SourceImage);
          Images.Add(DestImage)
        end
    end
end;

function TAdvancedIcon.LoadHeader(Stream: TStream; var IconCount: word): boolean;
var
  Header: TIconHeader;
begin
  Result:=false;
  with Stream do
    begin
      Read(Header,SizeOf(Header));
      if not ((Header.wReserved=0) and (Header.wType=1)) then exit;
      IconCount:=Header.wCount;
      Images.Capacity:=IconCount;
      Result:=True
    end
end;

function TAdvancedIcon.LoadFromResource(Instance: THandle; ResourceName: pChar): boolean;
Var
  Stream: TResourceStream;
  I: integer;
  Image: TIconImage;
  IconCount: word;
begin
  Result:=false;
  if IsBadStringPtr(ResourceName,63) then
    Stream:=TResourceStream.CreateFromID(Instance,integer(Resourcename),RT_Group_Icon)
  else
    Stream:=TResourceStream.Create(Instance,Resourcename,RT_Group_Icon);
  with Stream do
    try
      if not LoadHeader(Stream,IconCount) then exit;
      for I:=1 to IconCount do
        begin
          Image:=TResourceIconImage.Create(Stream,Instance);
          if Assigned(Image) then Images.add(Image)
        end;
    Result:=True
  finally
    Stream.Free
  end
end;

function TAdvancedIcon.LoadFromResourceID(Instance: THandle; ResourceID: integer): boolean;
begin
  Result:=LoadFromResource(Instance,MakeIntResource(ResourceID))
end;

function TAdvancedIcon.LoadFromResourceName(Instance: THandle; ResourceName: string): boolean;
begin
  Result:=LoadFromResource(Instance,PChar(ResourceName))
end;

function TAdvancedIcon.LoadFromFile(Filename: string): boolean;
var
  Stream: TFileStream;
  IconCount: word;
  I: integer;
  Image: TIconImage;
begin
  Result:=False;
  Images.Clear;
  Stream:=TFileStream.Create(Filename,fmOpenRead or fmShareDenyWrite);
  with Stream do
    try
      if not LoadHeader(Stream,IconCount) then exit;
      for I:=1 to IconCount do
        begin
          Image:=TFileIconImage.Create(Stream);
          if Assigned(Image) then Images.add(Image)
        end;
      Result:=True
    finally
      Free
    end
end;

procedure TAdvancedIcon.SaveToStream(Stream: TStream);
var
  Header: TIconHeader;
  ResInfo: TFileIconResInfo;
  I: integer;
  StartPos,StreamPos: DWord;
begin
  if Images.Count = 0 then exit;
  with Header do
    begin
      wReserved:=0; wType:=1;
      wCount:=Images.Count
    end;
  with Stream do
    begin
      StartPos:=Position;
      Write(Header,SizeOf(Header));
      for I:=1 to Images.Count do
        begin
          ResInfo.ResInfo:=Images.Image[I-1].Info;
          ResInfo.dwImageOffset:=0;
          Write(ResInfo,SizeOf(ResInfo))
        end;
      for I:=1 to Images.Count do
        begin
          StreamPos:=Position-StartPos;
          Position:=SizeOf(Header)+(I*SizeOf(ResInfo))-SizeOf(DWord);
          Write(StreamPos,SizeOf(StreamPos));
          Position:=StreamPos;
          Images.Image[I-1].WriteIconDataToStream(Stream)
        end
    end
end;

procedure TAdvancedIcon.SaveToFile(Filename: string);
var
  Stream: TMemoryStream;
begin
  Stream:=TMemoryStream.Create;
  try
    SaveToStream(Stream);
    if Stream.Size>0 then Stream.SaveToFile(Filename)
  finally
    Stream.Free
  end
end;

end.
