unit IconLibrary;

interface

uses Windows, Messages, SysUtils, Classes, Graphics, ShellAPI, IconTypes,
     IconImage, AdvancedIcon;

type
  TIconList = class;

  TIconLibrary = class
  private
    fFilename: string;
  protected
    fIcons: TIconList;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure LoadSelf;
    procedure LoadFromFile(Filename: string);
    procedure SaveToFile(Filename: string);
    property Icons: TIconList read fIcons;
    property Filename: string read fFilename;
  end;

  TIconList = class(TStringList)
  public
    procedure Clear; override;
    procedure Delete(Index: integer); override;
    procedure AddIcon(IconName: string; Icon: TAdvancedIcon);
  end;

  TICLReader = class
  private
  protected
    fFilename: string;
    fIcons: TStringList;
  public
    constructor Create(Filename: string); virtual;
    destructor Destroy; override;
    function Execute: boolean; virtual; abstract;
    property Icons: TStringList read fIcons;
  end;

  T32BitReader = class(TICLReader)
  private
  protected
    Instance: THandle;
    FreeLib: boolean;
    procedure IconResFound(ResName: pChar);
  public
    constructor Create(Filename: string); override;
    function Execute: boolean; override;
    function ExecuteSelf: boolean;
    destructor Destroy; override;
  end;

  T16BitReader = class(TICLReader)
  private
    fStream: TStream;
    fAlignShift: word;
    fWin16Offset: word;
    fResourceTableStart: DWord;
    fNamestableStart: DWord;
    fIconCount: word;
    fGroupIconCount: word;
    fICLName: string;
    fIconsRes: PNameRecArray;
    fGroupIconsRes: PNameRecArray;
  protected
    procedure ReadFileHeader;
    procedure ReadResourceTable;
    procedure ReadTypeInfo;
    procedure ReadResourceNames;
  public
    constructor Create(AFilename: string); override;
    destructor Destroy; override;
    function Execute: boolean; override;
    function GetResourceIndex(ID: word; GroupIcon: boolean): integer;
    procedure ReadResource(Index: word; GroupIcon: boolean; var Data: Pointer; var ResName: string);
    procedure ReadResourceStream(Index: word; GroupIcon: boolean; Stream: TStream; var ResName: string);
    property IconCount: word read fGroupIconCount;
  end;

  T16BitWriter = class
  private
    fShiftAlignOffs: DWord;
    fNamesTableOffs: DWord;
    fOfs_Icons: PDWordArray;
    fOfs_GIcons: PDWordArray;
    fIcons: integer;
    fIconLib: TIconLibrary;
    fAlignShift: word;
    fStream: TMemoryStream;
    Null: array[1..256] of char;
  protected
    procedure WriteShiftAlign;
    procedure UpdateTableEntry(EntryPos: DWord; StartOffset,EndOffset: DWord);
    procedure WriteHeader;
    procedure WriteResTable;
    procedure WriteNameTable;
    procedure WriteIconData;
    procedure Align;
    property Stream : TMemoryStream read FStream;
  public
    constructor Create(AIconLib: TIconLibrary);
    destructor Destroy; override;
    procedure SaveToFile(Filename: string);
    property IconLib: TIconLibrary read fIconLib;
  end;

  TICL16IconImage = class(TIconImage)
  private

  protected

  public
    constructor Create(AInfo: TIconResInfo; AIconData: Pointer);
  end;

implementation

{$R-}

type
  TICL16Icon = class(TAdvancedIcon)
  private
    FResName: string;
  protected
    procedure LoadICL16IconImage(Reader: T16BitReader;GroupIcon: TStream);
  public
    procedure LoadICL16Icon(Reader: T16BitReader; Index: integer);
    property ResName: string read fResName;
  end;

type
  EICLReadError = class (Exception);
  EICLWriteError = class (Exception);

const
  IMAGE_DOS_SIGNATURE = $5A4D; { MZ }
  IMAGE_IM_SIGNATURE = $4D49;  { IM }
  IMAGE_OS2_SIGNATURE = $454E; { NE }
  IMAGE_W32_SIGNATURE = $4550; { PE }
  IMAGE_NAMETABLE_ICL: array[0..3] of char = 'ICL'+#0;

  E_R_InvWin16Offs = 'Invalid Windows 16 Offset';
  E_R_InvWin16Header = 'Invalid Windows 16 Header';
  E_R_NotValid16Bit = 'ICL-   ';
  E_R_NoResources = '   ';
  E_R_NoNamesTable = '    ';
  E_R_IconResNotFound = '   ';
  E_R_ResTableError = '   ';
  E_W_NoIcons = '     ';
  E_W_TooManyIcons = '      32767 ';

procedure ICLReadError(const ErrMsg: string);
begin
  raise EICLReadError.Create(ErrMsg)
end;

procedure ICLWriteError(const ErrMsg: string);
begin
  raise EICLReadError.Create(ErrMsg)
end;

function IsValidLib(Filename: string; var Is32Bit: boolean): boolean;
var
  F: TFileStream;
  Sig: word;
begin
  Result:=False;
  F:=TFileStream.Create(Filename,fmOpenRead or fmShareDenyWrite);
  with F DO
    try
      Read(Sig,SizeOf(Sig));
      if (Sig <>IMAGE_DOS_SIGNATURE) then exit;
      Seek($3C,soFromBeginning);
      Read(Sig,SizeOf(Sig));
      Seek(Sig,soFromBeginning);
      Read(Sig,SizeOf(Sig));
      Result:=(Sig = IMAGE_W32_SIGNATURE) OR (Sig = IMAGE_OS2_SIGNATURE);
      Is32Bit:=(Sig = IMAGE_W32_SIGNATURE)
    finally
      F.Free
    end
end;

procedure TIconList.Clear;
var
  I: integer;
begin
  for I:=1 to Count do
    TAdvancedIcon(Objects[I-1]).Free;
  inherited
end;

procedure TIconList.Delete(Index: integer);
begin
  if not ((Index < 0) or (Index >= Count)) then
    TAdvancedIcon(Objects[Index]).Free;
  inherited
end;

procedure TIconList.AddIcon(IconName: string; Icon: TAdvancedIcon);
begin
  if not Assigned(Icon) then exit;
  if not (Icon is TAdvancedIcon) then exit;
  if Length(IconName)=0 then IconName:=Format('Icon %000d',[Count+1]);
  AddObject(IconName,Icon)
end;                      

constructor TIconLibrary.Create;
begin
  fIcons:=TIconList.Create
end;

destructor TIconLibrary.Destroy;
begin
  fIcons.Free
end;

procedure TIconLibrary.LoadSelf;
Var
  Reader: T32BitReader;
begin
  fFilename:=ParamStr(0);
  Reader:=T32BitReader.Create('');
  try
    if Reader.ExecuteSelf then
      Icons.Assign(Reader.Icons)
  finally
    Reader.Free
  end
end;

procedure TIconLibrary.LoadFromFile(Filename: string);
var
  LoadedIcons,IconCount: integer;
  Is32Bit: boolean;
  Reader: TICLReader;
begin
  LoadedIcons:=0;
  IconCount:=ExtractIcon(HInstance,PChar(Filename),dword(-1));
  if (IconCount=0) then exit;
  if not IsValidLib(Filename,Is32Bit) then exit;
  fFilename:=Filename;
  if not Is32Bit then
    Reader:=T16BitReader.Create(Filename)
  else
    Reader:=T32BitReader.Create(Filename);
  try
    if Reader.Execute then
      begin
        LoadedIcons:=Reader.Icons.Count;
        Icons.Assign(Reader.Icons)
      end
  finally
    Reader.Free
  end
end;

procedure TIconLibrary.SaveToFile(Filename: string);
Var
  ICLFile: T16BitWriter;
begin
  if Icons.Count = 0 Then ICLWriteError(E_W_NoIcons);
  if Icons.Count >= $8000 then ICLWriteError(E_W_TooManyIcons);
  ICLFile:=T16BitWriter.Create(self);
  try
    ICLFile.SaveToFile(Filename)
  finally
    ICLFile.Free
  end
end;

constructor TICLReader.Create(Filename: string);
begin
  fIcons:=TStringList.Create;
  fFilename:=Filename
end;

destructor TICLReader.Destroy;
begin
  if Assigned(fIcons) then fIcons.free;
  inherited
end;

function EnumResNameProc(hModule: THandle; lpszType,lpszName: PChar; lParam: lParam): bool; stdcall;
begin
  Result:=True;
  T32BitReader(lParam).IconResFound(lpszName)
end;

constructor T32BitReader.Create(Filename: string);
begin
  inherited;
  FreeLib:=False; Instance:=0
end;

function T32BitReader.Execute: boolean;
begin
  Result:=False;
  try
    Instance:=Windows.LoadLibraryEx(PChar(fFilename), 0, LOAD_LIBRARY_AS_DATAFILE)
  finally
  end;
  if Instance = 0 then exit;
  FreeLib:=True; Result:=True;
  EnumResourcenames(Instance,RT_GROUP_ICON,@EnumResNameProc,integer(self))
end;

function T32BitReader.ExecuteSelf: boolean;
begin
  EnumResourcenames(HInstance,RT_GROUP_ICON,@EnumResNameProc,integer(self));
  Result:=True
end;

procedure T32BitReader.IconResFound(ResName: pChar);
var
  Ico: TAdvancedIcon;
  IsID: boolean;
  RName: string;
begin
  Ico:=TAdvancedIcon.Create;
  Ico.LoadFromResource(Instance,ResName);
  IsID:=IsBadStringPtr(ResName,63);
  if IsID then
    RName:=inttostr(integer(Resname))
  else
    Rname:=Strpas(ResName);
  fIcons.AddObject(RName,Ico)
end;

destructor T32BitReader.Destroy;
begin
  inherited;
  if FreeLib then FreeLibrary(Instance)
end;

function T16BitReader.GetResourceIndex(ID : Word;GroupIcon : Boolean) : Integer;
var
  Res: PNameRecArray;
  I,IMax: integer;
begin
  Result:=-1;
  if GroupIcon then
    begin
      Res:=FGroupIconsRes;
      IMax:=FGroupIconCount
    end
  else
    begin
      Res:=FIconsRes;
      IMax:=FIconCount
    end;
  I:=0;
  while (I < IMax) do
    begin
      if ((Res^[I].rnID) xor $8000) = ID then
        begin
          Result:=I;
          break
        end;
      inc(I)
    end;
  if Result >= 0 then exit;
  ICLReadError(E_R_IconResNotFound)
end;

procedure T16BitReader.ReadResource(Index: word; GroupIcon: boolean; var Data: Pointer; var ResName: string);
var
  Namerec: TNameRec;
  ResStartPos: Longint;
  ResLength: Longint;
begin
  if GroupIcon then
    NameRec:=fGroupIconsRes^[Index]
  else
    NameRec:=fIconsRes^[Index];
  ResName:=Format('%s %0.3d',[fICLName,Index+1]);
  ResStartPos:=NameRec.rnOffset shl fAlignShift;
  ResLength:=NameRec.rnLength shl fAlignShift;
  fStream.Position:=ResStartPos;
  Getmem(Data,ResLength); fStream.Read(Data^,ResLength)
end;

procedure T16BitReader.ReadResourceStream(Index: word; GroupIcon: boolean; Stream: TStream; var ResName: string);
var
  Namerec: TNameRec;
  ResStartPos: longint;
  ResLength: longint;
begin
  if GroupIcon then
    NameRec:=fGroupIconsRes^[Index]
  else
    NameRec:=fIconsRes^[Index];
  ResName:=Format('%s %0.3d',[fICLName,Index+1]);
  ResStartPos:=NameRec.rnOffset shl fAlignShift;
  ResLength:=NameRec.rnLength shl fAlignShift;
  fStream.Position:=ResStartPos;
  Stream.CopyFrom(fStream,ResLength)
end;

constructor T16BitReader.Create(AFilename: string);
begin
  inherited;
  fStream:=TFileStream.Create(AFilename,fmOpenRead or fmShareDenyWrite);
  fStream.Position:=0; fICLName:=ExtractFilename(AFilename);
  Delete(fICLName,Length(fICLName)-Length(ExtractFileExt(fICLname))+1,30)
end;

destructor T16BitReader.Destroy;
begin
  fStream.Free;
  inherited
end;

function T16BitReader.Execute: boolean;
var
  ICLIcon: TICL16Icon;
  I,L: integer;
  S: string;
  NameOfList: boolean;
  SL: TStringList;
begin
  try
    fIconsRes:=nil; fGroupIconsRes:=nil;
    fIcons.Clear;
    ReadFileHeader; ReadResourceTable;
    SL:=TStringList.Create;
    try
      L:=fGroupIconCount;
      NameOfList:=(fIcons.Count=fGroupIconCount);
      for I:=1 to L do
        begin
          ICLIcon:=TICL16Icon.Create;
          ICLIcon.LoadICL16Icon(self,I-1);
          if NameOfList then
            S:=FIcons.Strings[I-1]
          else
            S:=ICLIcon.ResName;
          SL.AddObject(S,ICLIcon)
        end;
      fIcons.Assign(SL)
    finally
      SL.Free
    end
  except
    Result:=False;
    exit
  end;
  Result:=True
end;

procedure T16BitReader.ReadFileHeader;
var
  DosSignature, Win16Signature,
  ResTableOffset, NamesTableOffset: word;
begin
  with fStream do
    begin
      Read(DosSignature,SizeOf(DosSignature));
      if (DosSignature <> IMAGE_DOS_SIGNATURE) then ICLReadError(E_R_NotValid16Bit);
      Seek($3C,soFromBeginning);
      Read(fWin16Offset,SizeOf(Word));
      if (fWin16Offset < $3C) then ICLReadError(E_R_InvWin16Offs);
      Seek(fWin16Offset,soFromBeginning);
      Read(Win16Signature,SizeOf(Win16Signature));
      if (Win16Signature <> IMAGE_OS2_SIGNATURE) then ICLReadError(E_R_InvWin16Header);
      Seek($22,soFromCurrent);
      Read(ResTableOffset,SizeOf(Word));
      if (ResTableOffset = 0) then ICLReadError(E_R_NoResources);
      fResourceTableStart:=ResTableOffset+fWin16Offset;
      Read(NamesTableOffset,SizeOf(NamesTableOffset));
      if (NamesTableOffset = 0) then ICLReadError(E_R_NoNamesTable);
      fNamesTableStart:=NamesTableOffset+fWin16Offset
    end
end;

procedure T16BitReader.ReadResourceTable;
var
  EndTypes: word;
begin
  with FStream do
    begin
      Seek(FResourceTableStart,soFromBeginning);
      Read(fAlignShift,sizeof(word));
      Read(EndTypes,sizeof(word));
      while (Endtypes > 0) do
        begin
          Seek(-sizeof(word),sofromCurrent);
          ReadTypeInfo;
          Read(EndTypes,sizeof(word))
        end;
      if (fGroupIconsRes = nil) OR (fIconsRes = nil) then ICLReadError(E_R_ResTableError);
      ReadResourceNames
    end
end;

procedure T16BitReader.ReadResourceNames;
var
  NameLen: byte;
  S: string;
begin
  with fStream do
    begin
      Read(NameLen,sizeof(NameLen));
      while (NameLen>0) do
        begin
          SetString(S,nil,NameLen);
          Read(Pointer(S)^,NameLen);
          if not ((fIcons.Count = 0) AND (CompareStr(S,StrPas(IMAGE_NAMETABLE_ICL)) = 0)) then
            fIcons.Add(S);
          Read(NameLen,sizeof(NameLen))
        end
    end
end;

procedure T16BitReader.ReadTypeInfo;
var
  TypeInfo: TTypeRec;
  I: integer;
  TypeRead: boolean;
begin
  with fStream do
    begin
      Read(TypeInfo,sizeof(TypeInfo));
      with TypeInfo do
        begin
          i:=sizeof(TNameRec)*rtResourceCount;
          TypeRead:=false;
          if (rtTypeID or $8000 = rtTypeID) then
            begin
              if (Lo(rtTypeID) = word(RT_ICON)) then
                begin
                  fIconCount:=rtResourceCount;
                  GetMem(fIconsRes,I);
                  Read(fIconsRes^,I);
                  TypeRead:=True
                end;
              if (Lo(rtTypeID) = word(RT_GROUP_ICON)) then
                begin
                  fGroupIconCount:=rtResourceCount;
                  GetMem(fGroupIconsRes,I);
                  Read(fGroupIconsRes^,I);
                  TypeRead:=True
                end
            end
        end;
      if not TypeRead then Seek(I,soFromCurrent)
    end
end;

constructor T16BitWriter.Create(AIconLib: TIconLibrary);
var
  I: integer;
begin
  fIconLib:=AIconLib; fIcons:=0;
  fStream:=TMemoryStream.Create;
  for I:=1 to IconLib.Icons.Count do
    inc(fIcons,TAdvancedIcon(IconLib.Icons.Objects[I-1]).Images.Count);
  fAlignShift:=1;
  GetMem(fOfs_Icons,fIcons*sizeof(DWord));
  GetMem(fOfs_GIcons,IconLib.Icons.Count*sizeof(DWord));
  FillChar(Null,sizeof(Null),#0)
end;

procedure T16BitWriter.WriteShiftAlign;
var
  I,J: integer;
  Ico: TAdvancedIcon;
  IconDataSize: DWord;
  SavedPos: Cardinal;
begin
  IconDataSize:=Stream.Position;
  for I:=1 to IconLib.Icons.Count do
    begin
      Ico:=TAdvancedIcon(IconLib.Icons.Objects[I-1]);
      inc(fIcons,Ico.Images.Count);
      inc(IconDataSize,sizeof(TIconHeader));
      for J:=1 to Ico.Images.Count do
        begin
          inc(IconDataSize,sizeof(TResourceIconResInfo));
          inc(IconDataSize,Ico.Images.Image[J-1].Info.BytesInRes)
        end
    end;
  fAlignShift:=0;
  while ((MaxWord shl fAlignShift) < IconDataSize) do inc(fAlignShift);
  with Stream do
    begin
      SavedPos:=Position; Position:=fShiftAlignOffs;
      Write(fAlignShift,SizeOf(fAlignShift));
      Position:=SavedPos
    end
end;

destructor T16BitWriter.Destroy;
begin
  FStream.Free;
  Freemem(fOfs_Icons); Freemem(fOfs_GIcons);
  inherited
end;

procedure T16BitWriter.SaveToFile(Filename: string);
begin
  try
    WriteHeader;
    WriteResTable; WriteNameTable;
    WriteShiftAlign; WriteIconData;
    fStream.SaveToFile(Filename)
  finally
  end
end;

procedure T16BitWriter.WriteHeader;
var
  W: word;
  Null: array[1..100] of Char;
begin
  FillChar(Null,sizeof(Null),#0);
  with Stream do
    begin
      W:=IMAGE_DOS_SIGNATURE;
      write(W,sizeof(W));
      write(Null,58); W:=$80;
      write(W,sizeof(W));
      write(Null,64);
      W:=IMAGE_IM_SIGNATURE;
      write(W,sizeof(W));
      W:=IMAGE_OS2_SIGNATURE;
      write(W,sizeof(W));
      write(Null,32); W:=64;
      write(W,sizeof(W));
      write(W,sizeof(W));
      fNamesTableOffs:=Position;
      write(Null,16); W:=$2;
      write(W,sizeof(W));
      write(Null,7); W:=$3;
      write(W,1)
    end
end;

procedure T16BitWriter.WriteResTable;
var
  W: word;
  DW: DWord;
  I: integer;
  NameInfo: TNameRec;
begin
  DW:=0; W:=0;
  with Stream do
    begin
      fShiftAlignOffs:=Position;
      write(W,sizeof(W)); W:=$800E;
      write(W,sizeof(W));
      W:=IconLib.Icons.Count;
      write(W,sizeof(W));
      write(DW,sizeof(DW));
      FillChar(NameInfo,sizeof(NameInfo),#0);
      NameInfo.rnFlags:=$1C30;
      NameInfo.rnOffset:=$FF;
      NameInfo.rnLength:=$FF00;
      for I:=1 to IconLib.Icons.Count do
        begin
          NameInfo.rnID:=I or $8000;
          fOfs_GIcons^[I-1]:=Position;
          write(NameInfo,sizeof(NameInfo))
       end;
      W:=$8003; write(W,sizeof(W));
      W:=fIcons; write(W,sizeof(W));
      Write(DW,SizeOf(DW));
      for I:=1 to fIcons do
        begin
          NameInfo.rnID:=I or $8000;
          fOfs_Icons^[I-1]:=Position;
          write(NameInfo,sizeof(NameInfo))
        end;
      W:=0; write(W,SizeOf(W))
    end
end;

procedure T16BitWriter.WriteNameTable;
var
  B: byte;
  I,Pos: integer;
  S: string;
  W: word;
begin
  with Stream do
    begin
      Pos:=Position;
      Position:=fNamesTableOffs;
      W:=Pos-$80;
      write(W,sizeof(word));
      Position:=Pos; B:=3;
      write(B,sizeof(B));
      write(IMAGE_NAMETABLE_ICL,3);
      for I:=1 to IconLib.Icons.Count do
        begin
          S:=IconLib.Icons.Strings[I-1];
          B:=length(S);
          write(B,sizeof(B)); write(Pointer(S)^,B)
        end;
      B:=0; write(B,sizeof(B))
    end
end;

procedure T16BitWriter.Align;
var
  C: integer;
begin
  with Stream do
    begin
      C:=(Position shr fAlignShift);
      if (C shl fAlignShift) = Position then exit;
      inc(C); C:=(C shl fAlignShift)-Position;
      if C > sizeof(Null) then
        begin
          while (C > sizeof(Null)) do
            begin
              write(Null,sizeof(Null));
              dec(C,sizeof(Null))
            end
        end;
      write(Null,C)
    end
end;

procedure T16BitWriter.WriteIconData;
var
  IID,W: Word;
  I,J: Integer;
  Ico: TAdvancedIcon;
  Header: TIconHeader;
  ResInfo: TResourceIconResInfo;
  SOffset,EOffset: DWord;
begin
  Header.wReserved:=0; Header.wType:=1;
  IID:=1; Align;
  with Stream do
    begin
      for I:=1 to IconLib.Icons.Count do
        begin
          SOffset:=Position;
          Ico:=TAdvancedIcon(IconLib.Icons.Objects[I-1]);
          Header.wCount:=Ico.Images.Count;
          write(Header,SizeOf(Header));
          for J:=1 to Header.wCount do
            begin
              ResInfo.ResInfo:=Ico.Images.Image[J-1].Info;
              ResInfo.ID:=IID;
              write(ResInfo,SizeOf(ResInfo)); inc(IID)
            end;
          Align; EOffset:=Position;
          UpdateTableEntry(fOfs_GIcons^[I-1],SOffset,EOffset)
        end;
      IID:=1;
      for I:=1 to IconLib.Icons.Count do
        begin
          Ico:=TAdvancedIcon(IconLib.Icons.Objects[I-1]);
          for J:=1 to Ico.Images.Count do
            begin
              SOffset:=Position;
              Ico.Images.Image[J-1].WriteIconDataToStream(Stream);
              Align; EOffset:=Position;
              UpdateTableEntry(fOfs_Icons^[IID-1],SOffset,EOffset);
              inc(IID)
            end
        end;
      W:=0; write(W,SizeOf(W))
    end
end;

procedure T16BitWriter.UpdateTableEntry(EntryPos : DWord;StartOffset,EndOffset : DWord);
var
  OldPos: cardinal;
  W: word;
begin
  with Stream do
    begin
      OldPos:=Position;
      try
        Position:=EntryPos;
        W:=StartOffset shr fAlignShift;
        write(W,sizeof(word));
        W:=(EndOffset-StartOffset) shr fAlignShift;
        write(W,sizeof(word));
      finally
        Position:=OldPos
      end
    end
end;

constructor TICL16IconImage.Create(AInfo: TIconResInfo; AIconData: Pointer);
begin
  inherited Create;
  fInfo:=AInfo; fIconData:=AIconData
end;

procedure TICL16Icon.LoadICL16Icon(Reader: T16BitReader; Index: integer);
var
  GroupIcon: TMemoryStream;
  IconCount: word;
  I: integer;
begin
  GroupIcon:=TMemoryStream.Create;
  try
    Reader.ReadResourceStream(Index,True,GroupIcon,FResname);
    GroupIcon.Position:=0;
    if not LoadHeader(GroupIcon,IconCount) then exit;
    for I:=1 to IconCount do LoadICL16IconImage(Reader,GroupIcon)
  finally
    GroupIcon.Free
  end
end;

procedure TICL16Icon.LoadICL16IconImage(Reader: T16BitReader; GroupIcon: TStream);
var
  Info: TResourceIconResInfo;
  Image: TICL16IconImage;
  Index: integer;
  S: string;
  Data: Pointer;
begin
  GroupIcon.Read(Info,SizeOf(Info));
  Index:=Reader.GetResourceIndex(Info.ID,False);
  if Index<0 then exit;
  Reader.ReadResource(Index,False,Data,S);
  Image:=TICL16IconImage.Create(Info.ResInfo,Data);
  Images.Add(Image)
end;

initialization

end.
